@@ -1,4 +1,22 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2012-2017 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 """ Keyboard layout view """
 
 from __future__ import division, print_function, unicode_literals
@@ -7,6 +25,8 @@ import time
 from math import pi
 
 import cairo
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 from gi.repository         import Gtk, Gdk, GdkPixbuf
 
 from Onboard.utils         import Rect, \
@@ -16,8 +36,8 @@ from Onboard.utils         import Rect, \
 from Onboard.WindowUtils   import get_monitor_dimensions
 from Onboard.KeyGtk        import Key
 from Onboard.KeyCommon     import LOD
+from Onboard.definitions   import UIMask
 
-import Onboard.Keyboard
 
 ### Logging ###
 import logging
@@ -69,12 +89,13 @@ class LayoutView:
     def get_color_scheme(self):
         return self.keyboard.color_scheme
 
-    def invalidate_for_resize(self, lod = LOD.FULL):
+    def invalidate_for_resize(self, lod=LOD.FULL):
         self.invalidate_keys()
         if self._lod == LOD.FULL:
             self.invalidate_shadows()
         self.invalidate_font_sizes()
-        #self.invalidate_label_extents()
+        # self.invalidate_label_extents()
+        self.keyboard.invalidate_for_resize()
 
     def invalidate_font_sizes(self):
         """
@@ -145,7 +166,7 @@ class LayoutView:
     def raise_to_top(self):
         pass
 
-    def redraw(self, items = None, invalidate = True):
+    def redraw(self, items=None, invalidate=True):
         """
         Queue redrawing for individual keys or the whole keyboard.
         """
@@ -174,7 +195,7 @@ class LayoutView:
 
             self.queue_draw_area(*area)
 
-    def redraw_labels(self, invalidate = True):
+    def redraw_labels(self, invalidate=True):
         self.redraw(self.update_labels(), invalidate)
 
     def update_transparency(self):
@@ -203,7 +224,7 @@ class LayoutView:
         pass
 
     def apply_ui_updates(self, mask):
-        if mask & Onboard.Keyboard.UIMask.SIZE:
+        if mask & UIMask.SIZE:
             self.invalidate_for_resize()
 
     def update_layout(self):
@@ -305,8 +326,7 @@ class LayoutView:
         transparent_bg = False
         plain_bg = False
 
-        if config.xid_mode and \
-           config.is_keep_aspect_ratio_enabled():
+        if config.is_keep_xembed_frame_aspect_ratio_enabled():
             if self.supports_alpha:
                 self._clear_xembed_background(context)
                 transparent_bg = True
@@ -384,7 +404,8 @@ class LayoutView:
         except AttributeError:
             size, size_mm = get_monitor_dimensions(self)
             filename = config.get_desktop_background_filename()
-            if not filename:
+            if not filename or \
+               size[0] <= 0 or size[1] <= 0:
                 pixbuf = None
             else:
                 try:
@@ -416,6 +437,9 @@ class LayoutView:
         rect = self.get_keyboard_frame_rect()
         fill = self.get_background_rgba()
 
+        if self.can_draw_sidebars():
+            self._draw_side_bars(context)
+
         fill_gradient = config.theme_settings.background_gradient
         if lod == LOD.MINIMAL or \
            fill_gradient == 0:
@@ -442,16 +466,48 @@ class LayoutView:
             roundrect_arc(context, rect, corner_radius)
         else:
             context.rectangle(*rect)
+
         context.fill()
 
         if frame:
             self.draw_window_frame(context, lod)
             self.draw_keyboard_frame(context, lod)
 
+    def _draw_side_bars(self, context):
+        """
+        Transparent bars left and right of the aspect corrected
+        keyboard frame.
+        """
+        rgba = self.get_background_rgba()
+        rgba[3] = 0.5
+        rwin = Rect(0, 0,
+                    self.get_allocated_width(),
+                    self.get_allocated_height())
+        rframe = self.get_keyboard_frame_rect()
+
+        if rwin.w > rframe.w:
+            r = rframe.copy()
+            context.set_source_rgba(*rgba)
+            context.set_line_width(0)
+
+            r.x = rwin.left()
+            r.w = rframe.left() - rwin.left()
+            context.rectangle(*r)
+            context.fill()
+
+            r.x = rframe.right()
+            r.w = rwin.right() - rframe.right()
+            context.rectangle(*r)
+            context.fill()
+
     def can_draw_frame(self):
-        """ overloaded in popups """
+        """ overloaded in KeyboardWidget """
         return True
 
+    def can_draw_sidebars(self):
+        """ overloaded in KeyboardWidget """
+        return False
+
     def draw_window_frame(self, context, lod):
         pass
 
@@ -527,12 +583,16 @@ class LayoutView:
         """
         Draw drop shadows for all keys.
         """
+        # Shadows are drawn at odd positions when resizing while
+        # docked and extended with side bars visible.
+        # -> Turn them off while resizing. Improves rendering speed a bit too.
+        if lod < LOD.FULL:
+            return
         if not config.theme_settings.key_shadow_strength:
             return
 
         self._auto_select_shadow_quality(context)
 
-        # draw shadows
         context.save()
         self.set_shadow_scale(context, lod)
 
@@ -663,7 +723,7 @@ class LayoutView:
         else:
             rect = Rect(0, 0, self.get_allocated_width(),
                               self.get_allocated_height())
-        return rect
+        return rect.int()
 
     def is_docking_expanded(self):
         return self.window.docking_enabled and self.window.docking_expanded
@@ -671,7 +731,7 @@ class LayoutView:
 
     def update_labels(self, lod = LOD.FULL):
         """
-        Iterates through all key groups and set each key's
+        Iterate through all key groups and set each key's
         label font size to the maximum possible for that group.
         """
         changed_keys = set()
@@ -680,7 +740,7 @@ class LayoutView:
         mod_mask = self.keyboard.get_mod_mask()
 
         if layout:
-            if lod == LOD.FULL: # no label changes necessary while dragging
+            if lod == LOD.FULL:  # no label changes necessary while dragging
 
                 for key in layout.iter_keys():
                     old_label = key.get_label()
@@ -693,11 +753,17 @@ class LayoutView:
                 for key in keys:
                     best_size = key.get_best_font_size(mod_mask)
                     if best_size:
-                        if not max_size or best_size < max_size:
-                            max_size = best_size
+                        if key.ignore_group:
+                            if key.font_size != best_size:
+                                key.font_size = best_size
+                                changed_keys.add(key)
+                        else:
+                            if not max_size or best_size < max_size:
+                                max_size = best_size
 
                 for key in keys:
-                    if key.font_size != max_size:
+                    if key.font_size != max_size and \
+                       not key.ignore_group:
                         key.font_size = max_size
                         changed_keys.add(key)
 
@@ -711,3 +777,20 @@ class LayoutView:
             return layout.get_key_at(point, keyboard.active_layer)
         return None
 
+    def get_xid(self):
+        # Zesty, X, Gtk 3.22: XInput select_events() on self leads to
+        # LP: #1636252. On the first call to get_xid() of a child widget,
+        # Gtk creates a new native X Window with broken transparency.
+        # The toplevel window ought to always have a native X window, so
+        # we'll pick that one instead and skip on-the fly creation.
+        # TouchInput isn't used for anything other than full client areas
+        # yet, so in principle this shouldn't be a problem.
+
+        toplevel = self.get_toplevel()
+        if toplevel:
+            topwin = toplevel.get_window()
+            if topwin:
+                return topwin.get_xid()
+        return 0
+
+
