@@ -1,4 +1,9 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2012-2017 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
 # Onboard is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 3 of the License, or
@@ -11,10 +16,6 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
-# Copyright © 2012, marmuta
-#
-# This file is part of Onboard.
 
 from __future__ import division, print_function, unicode_literals
 
@@ -23,8 +24,9 @@ import re
 import glob
 
 import logging
-_logger = logging.getLogger("TextDomain")
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 try:
     from gi.repository import Atspi
 except ImportError as e:
@@ -33,6 +35,7 @@ except ImportError as e:
 from Onboard.TextChanges  import TextSpan
 from Onboard.utils        import KeyCode, Modifiers, unicode_str
 
+_logger = logging.getLogger("TextDomain")
 
 class TextDomains:
     """ Collection of all recognized text domains. """
@@ -58,7 +61,9 @@ class TextDomains:
 
 
 class TextDomain:
-    """ Abstract base class as a catch-all for domain specific functionalty. """
+    """
+    Abstract base class as a catch-all for domain specific functionalty.
+    """
 
     def __init__(self):
         self._url_parser = PartialURLParser()
@@ -72,7 +77,7 @@ class TextDomain:
         """ Called on being selected as the currently active domain. """
         pass
 
-    def read_context(self, accessible):
+    def read_context(self, keyboard, accessible):
         return NotImplementedError()
 
     def get_text_begin_marker(self):
@@ -98,17 +103,25 @@ class TextDomain:
         ''
 
         # filename
+        >>> import tempfile
+        >>> from os.path import abspath, dirname, join
+        >>> def touch(fn):
+        ...     with open(fn, mode="w") as f: n = f.write("")
+        >>> td = tempfile.TemporaryDirectory(prefix="test_onboard _")
+        >>> dir = td.name
+
+        >>> touch(join(dir, "onboard-defaults.conf.example"))
+        >>> touch(join(dir, "onboard-defaults.conf.example.another"))
+        >>>
         >>> d.get_auto_separator("/etc")
         '/'
-        >>> from os.path import abspath, dirname
-        >>> path = dirname(abspath("./onboard-defaults.conf.example"))
-        >>> d.get_auto_separator(join(path, "onboard-defaults"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults"))
         '.'
-        >>> d.get_auto_separator(join(path, "onboard-defaults.conf"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults.conf"))
         '.'
-        >>> d.get_auto_separator(join(path, "onboard-defaults.conf.example"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults.conf.example"))
         ''
-        >>> d.get_auto_separator(join(path, "onboard-defaults.conf.example.nexus7"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults.conf.example.another"))
         ' '
 
         # filename in directory with spaces
@@ -116,19 +129,19 @@ class TextDomain:
         >>> td = tempfile.TemporaryDirectory(prefix="test onboard _")
         >>> dir = td.name
         >>> fn = os.path.join(dir, "onboard-defaults.conf.example")
-        >>> with open(fn, mode="w") as f: n = f.write("")
+        >>> touch(fn)
         >>> d.get_auto_separator(join(dir, "onboard-defaults"))
         '.'
 
         # no false positives for valid filenames before the current token
-        >>> d.get_auto_separator(path + "/onboard-defaults no-file")
+        >>> d.get_auto_separator(dir + "/onboard-defaults no-file")
         ' '
         """
         separator = " "
 
         # Split at whitespace to catch whole URLs/file names and
         # keep separators.
-        strings = re.split('(\s+)', context)
+        strings = re.split(r'(\s+)', context)
         if strings:
             string = strings[-1]
             if self._url_parser.is_maybe_url(string):
@@ -145,7 +158,7 @@ class TextDomain:
 
 
     def _search_valid_file_name(self, strings):
-        """
+        r"""
         Search for a valid filename backwards across separators.
 
         Doctests:
@@ -161,17 +174,17 @@ class TextDomain:
         >>> with open(fn2, mode="w") as f: n = f.write("")
 
         # simple file in dir with spaces must return as filename
-        >>> strings = re.split('(\s+)', fn1)
+        >>> strings = re.split(r'(\s+)', fn1)
         >>> "/test onboard" in d._search_valid_file_name(strings)
         True
 
         # file with spaces in dir with spaces must return as filename
-        >>> strings = re.split('(\s+)', fn2)
+        >>> strings = re.split(r'(\s+)', fn2)
         >>> "/test onboard" in d._search_valid_file_name(strings)
         True
 
         # random string after a valid file must not be confused with a filename
-        >>> strings = re.split('(\s+)', fn2 + " no-file")
+        >>> strings = re.split(r'(\s+)', fn2 + " no-file")
         >>> d._search_valid_file_name(strings) is None
         True
         """
@@ -203,19 +216,19 @@ class TextDomain:
         Doctests:
         >>> d = DomainGenericText()
 
-        # Span doesn't change for simple words
+        # Span doesn't grow for simple words
         >>> d.grow_learning_span(TextSpan(8, 1, "word1 word2 word3"))
         (8, 1, 'r')
 
-        # Span growing to cover a complete URL
+        # Span grows to include a complete URL
         >>> d.grow_learning_span(TextSpan(13, 1, "http://www.domain.org"))
         (0, 21, 'http://www.domain.org')
 
-        # Span growing to cover multiple complete URLs
+        # Span grows to include multiple complete URLs
         >>> d.grow_learning_span(TextSpan(19, 13, "http://www.domain.org word http://slashdot.org"))
         (0, 46, 'http://www.domain.org word http://slashdot.org')
 
-        # Span growing to cover a complete filename
+        # Span grows to include a complete filename
         >>> d.grow_learning_span(TextSpan(10, 1, "word1 /usr/bin/bash word2"))
         (6, 13, '/usr/bin/bash')
 
@@ -229,8 +242,9 @@ class TextDomain:
         >>> d.grow_learning_span(TextSpan(18, 1, "word1 /usr/bin/bash word2"))
         (6, 13, '/usr/bin/bash')
 
-        # Large text with text offset > 0: returned position must be offset too.
-        >>> d.grow_learning_span(TextSpan(116, 1, "word1 /usr/bin/bash word2", 100))
+        # Large text with text offset>0: returned position must be offset too
+        >>> d.grow_learning_span(TextSpan(116, 1,
+        ...     "word1 /usr/bin/bash word2", 100))
         (106, 13, '/usr/bin/bash')
         """
         text   = text_span.get_text()
@@ -238,19 +252,19 @@ class TextDomain:
         begin  = text_span.begin() - offset
         end    = text_span.end() - offset
 
-        tokens, spans = self._split_growth_sections(text)
+        sections, spans = self._split_growth_sections(text)
 
         for i, s in enumerate(spans):
-            if begin < s[1] and end > s[0]: # intersects?
+            if begin < s[1] and end > s[0]:  # intersects?
 
-                token = tokens[i]
+                section = sections[i]
                 span = spans[i]
-                if self._url_parser.is_maybe_url(token) or \
-                   self._url_parser._is_maybe_filename(token):
+                if self._url_parser.is_maybe_url(section) or \
+                   self._url_parser._is_maybe_filename(section):
                     begin = min(begin, span[0])
                     end = max(end, span[1])
 
-        return begin+offset, end - begin, text[begin:end]
+        return begin + offset, end - begin, text[begin:end]
 
     def can_record_insertion(self, accessible, pos, length):
         return True
@@ -258,10 +272,13 @@ class TextDomain:
     def can_give_keypress_feedback(self):
         return True
 
-    def can_spell_check(self):
+    def can_spell_check(self, section_span):
         return False
 
-    def can_auto_correct(self):
+    def can_auto_correct(self, section_span):
+        return False
+
+    def can_auto_punctuate(self, has_begin_of_text):
         return False
 
     def can_suggest_before_typing(self):
@@ -271,7 +288,7 @@ class TextDomain:
     def handle_key_press(self, keycode, mod_mask):
         return True, None  # entering_text, end_of_editing
 
-    _growth_tokens_pattern = re.compile("[^\s?#@]+", re.DOTALL)
+    _growth_sections_pattern = re.compile(r"[^\s?#@]+", re.DOTALL)
 
     def _split_growth_sections(self, text):
         """
@@ -289,7 +306,7 @@ class TextDomain:
         >>> d._split_growth_sections("http://user:pass@www.domain.org")
         (['http://user:pass', 'www.domain.org'], [(0, 16), (17, 31)])
         """
-        matches = self._growth_tokens_pattern.finditer(text)
+        matches = self._growth_sections_pattern.finditer(text)
         tokens = []
         spans = []
         for m in matches:
@@ -304,7 +321,7 @@ class DomainNOP(TextDomain):
     def matches(self, **kwargs):
         return True
 
-    def read_context(self, accessible):
+    def read_context(self, keyboard, accessible):
         return "", "", 0, TextSpan(), False, 0
 
     def get_auto_separator(self, context):
@@ -328,60 +345,60 @@ class DomainGenericText(TextDomain):
     def matches(self, **kwargs):
         return TextDomain.matches(self, **kwargs)
 
-    def read_context(self, accessible):
+    def read_context(self, keyboard, accessible):
         """ Extract prediction context from the accessible """
 
         # get caret position from selection
-        selection = None
-        try:
-            sel = accessible.get_selection(0)
-            # Gtk-2 applications return 0,0 when there is no selection.
-            # Gtk-3 applications return caret positions in that case.
-            # LibreOffice Writer in Vivid initially returns -1,-1 when there
-            # is no selection, later the caret position.
-            start = sel.start_offset
-            end = sel.end_offset
-            if start > 0 and \
-               end > 0 and \
-               start <= end:
-                selection = (sel.start_offset, sel.end_offset)
-        except Exception as ex: # Private exception gi._glib.GErro
-            _logger.info("DomainGenericText.read_context(), selection: " \
-                         + unicode_str(ex))
+        selection = accessible.get_selection()
 
         # get text around the caret position
         try:
+            count = accessible.get_character_count()
+
             if selection is None:
                 offset = accessible.get_caret_offset()
+
+                # In Zesty, firefox 50.1 often returns caret position -1
+                # when typing into the urlbar. Assume we are at the end
+                # of the text when that happens.
+                if offset < 0:
+                    _logger.warning("DomainGenericText.read_context(): "
+                                    "Atspi.Text.get_caret_offset() "
+                                    "returned invalid {}. "
+                                    "Pretending the cursor is at the end "
+                                    "of the text at offset {}."
+                                    .format(offset, count))
+                    offset = count
+
                 selection = (offset, offset)
-            r = accessible.get_text_at_offset(selection[0],
-                                Atspi.TextBoundaryType.LINE_START)
-            count = accessible.get_character_count()
-        except Exception as ex: # Private exception gi._glib.GError when
-                                # gedit became unresponsive.
-            _logger.info("DomainGenericText.read_context(), text: " \
-                         + unicode_str(ex))
+
+            r = accessible.get_text_at_offset(
+                selection[0], Atspi.TextBoundaryType.LINE_START)
+        except Exception as ex:     # Private exception gi._glib.GError when
+                                    # gedit became unresponsive.
+            _logger.info("DomainGenericText.read_context(), text: " +
+                         unicode_str(ex))
             return None
 
-        line = unicode_str(r.content).replace("\n","")
+        line = unicode_str(r.content).replace("\n", "")
         line_caret = max(selection[0] - r.start_offset, 0)
 
         begin = max(selection[0] - 256, 0)
         end   = min(selection[0] + 100, count)
         try:
-            text = Atspi.Text.get_text(accessible, begin, end)
-        except Exception as ex: # Private exception gi._glib.GError when
-                                # gedit became unresponsive.
-            _logger.info("DomainGenericText.read_context(), text2: " \
-                         + unicode_str(ex))
+            text = accessible.get_text(begin, end)
+        except Exception as ex:     # Private exception gi._glib.GError when
+                                    # gedit became unresponsive.
+            _logger.info("DomainGenericText.read_context(), text2: " +
+                         unicode_str(ex))
             return None
 
         text = unicode_str(text)
 
         # Not all text may be available for large selections. We only need the
         # part before the begin of the selection/caret.
-        selection_span = TextSpan(selection[0], selection[1]-selection[0],
-                              text, begin)
+        selection_span = TextSpan(selection[0], selection[1] - selection[0],
+                                  text, begin)
         context = text[:selection[0] - begin]
         begin_of_text = begin == 0
         begin_of_text_offset = 0
@@ -389,10 +406,31 @@ class DomainGenericText(TextDomain):
         return (context, line, line_caret, selection_span,
                 begin_of_text, begin_of_text_offset)
 
-    def can_spell_check(self):
+    def can_spell_check(self, section_span):
+        """
+        Can we auto-correct this span?.
+
+        Doctests:
+        >>> d = DomainGenericText()
+        >>> d.can_spell_check(TextSpan(0, 3, "abc"))
+        True
+        >>> d.can_spell_check(TextSpan(4, 3, "abc def"))
+        True
+        >>> d.can_spell_check(TextSpan(0, 20, "http://www.domain.org"))
+        False
+        """
+        section = section_span.get_span_text()
+        return not self._url_parser.is_maybe_url(section) and \
+               not self._url_parser._is_maybe_filename(section)
         return True
 
-    def can_auto_correct(self):
+    def can_auto_correct(self, section_span):
+        """
+        Can we auto-correct this span?.
+        """
+        return self.can_spell_check(section_span)
+
+    def can_auto_punctuate(self, has_begin_of_text):
         return True
 
     def get_text_begin_marker(self):
@@ -402,22 +440,24 @@ class DomainGenericText(TextDomain):
 class DomainTerminal(TextDomain):
     """ Terminal entry, in particular gnome-terminal """
 
-    _prompt_patterns = tuple(re.compile(p, re.UNICODE) for p in \
-                             ("^gdb$ ",
-                              "^>>> ", # python
-                              "^In \[[0-9]*\]: ",   # ipython
-                              "^:",    # vi command mode
-                              "^/",    # vi search
-                              "^\?",   # vi reverse search
-                              "\$ ",   # generic prompt
-                              "# ",    # root prompt
-                             )
+    _prompt_patterns = tuple(re.compile(p, re.UNICODE) for p in
+                                (
+                                    "^gdb$ ",
+                                    "^>>> ",              # python
+                                    r"^In \[[0-9]*\]: ",  # ipython
+                                    "^:",                 # vi command mode
+                                    "^/",                 # vi search
+                                    r"^\?",               # vi reverse search
+                                    r"\$ ",               # generic prompt
+                                    "# ",                 # root prompt
+                                    "^.*?@.*?/.*?> "      # fish
+                                )
                             )
 
-    _prompt_blacklist_patterns = tuple(re.compile(p, re.UNICODE) for p in \
-                             (
-                              "^\(.*\)`.*': ", # bash incremental search
-                             )
+    _prompt_blacklist_patterns = tuple(re.compile(p, re.UNICODE) for p in
+                                (
+                                    r"^\(.*\)`.*': ",  # bash incremental search
+                                )
                             )
 
     def matches(self, **kwargs):
@@ -427,21 +467,22 @@ class DomainTerminal(TextDomain):
     def init_domain(self):
         pass
 
-    def read_context(self, accessible, offset = None):
+    def read_context(self, keyboard, accessible):
         """
         Extract prediction context from the accessible
         """
-        if offset is None:
-            try:
-                offset = accessible.get_caret_offset()
-            except Exception as ex: # Private exception gi._glib.GError when
-                                    # gedit became unresponsive.
-                _logger.info("DomainTerminal.read_context(): " \
-                             + unicode_str(ex))
-                return None
+        try:
+            offset = accessible.get_caret_offset()
+        except Exception as ex:     # Private exception gi._glib.GError
+                                    # when gedit became unresponsive.
+            _logger.info("DomainTerminal.read_context(): " +
+                         unicode_str(ex))
+            return None
 
         context_lines, prompt_length, line, line_start, line_caret = \
-                             self._get_text_after_prompt(accessible, offset)
+            self._get_text_after_prompt(
+                accessible, offset,
+                keyboard.get_last_typed_was_separator())
 
         if prompt_length:
             begin_of_text = True
@@ -454,13 +495,14 @@ class DomainTerminal(TextDomain):
         before_line = "".join(context_lines[:-1])
         selection_span = TextSpan(offset, 0,
                                   before_line + line,
-                                  line_start-len(before_line))
+                                  line_start - len(before_line))
 
         result = (context, line, line_caret, selection_span,
                   begin_of_text, begin_of_text_offset)
         return result
 
-    def _get_text_after_prompt(self, accessible, caret_offset):
+    def _get_text_after_prompt(self, accessible, caret_offset,
+                               last_typed_was_separator=None):
         """
         Return text from the input area of the terminal after the prompt.
 
@@ -480,6 +522,8 @@ class DomainTerminal(TextDomain):
         ...         return r
         ...     def get_text_before_offset(self, offset, boundary):
         ...         return self.get_text_at_offset(offset - self._width, boundary)
+        ...     def is_byobu(self):
+        ...         return False
 
         >>> d = DomainTerminal()
 
@@ -512,16 +556,25 @@ class DomainTerminal(TextDomain):
 
         """
 
-        r = accessible.get_text_at_offset(caret_offset,
-                            Atspi.TextBoundaryType.LINE_START)
+        r = accessible.get_text_at_offset(
+            caret_offset, Atspi.TextBoundaryType.LINE_START)
         line = unicode_str(r.content)
         line_start = r.start_offset
         line_caret = caret_offset - line_start
-
         # remove prompt from the current or previous lines
         context_lines = []
         prompt_length = None
         l = line[:line_caret]
+
+        # Zesty: byobu running in gnome-terminal doesn't report trailing
+        # spaces in text and caret-position.
+        # Awful hack: assume there is always a trailing space when the caret
+        # is at the end of the line and we just typed a separator.
+        if line[line_caret:] == "\n" and \
+           last_typed_was_separator and \
+           accessible.is_byobu():
+            l += " "
+
         for i in range(2):
 
             # matching blacklisted prompt? -> cancel whole context
@@ -533,7 +586,7 @@ class DomainTerminal(TextDomain):
             prompt_length = self._find_prompt(l)
             context_lines.insert(0, l[prompt_length:])
             if i == 0:
-                line = line[prompt_length:] # cut prompt from input line
+                line = line[prompt_length:]  # cut prompt from input line
                 line_start += prompt_length
                 line_caret -= prompt_length
             if prompt_length:
@@ -541,8 +594,8 @@ class DomainTerminal(TextDomain):
 
             # no prompt yet -> let context reach
             # across one more line break
-            r = accessible.get_text_before_offset(caret_offset,
-                                Atspi.TextBoundaryType.LINE_START)
+            r = accessible.get_text_before_offset(
+                caret_offset, Atspi.TextBoundaryType.LINE_START)
             l = unicode_str(r.content)
 
         result = (context_lines, prompt_length,
@@ -575,7 +628,7 @@ class DomainTerminal(TextDomain):
         # Problem: learning won't happen for uncommon prompts, but less random
         # junk scrolling by should enter the user model in return.
         context_lines, prompt_length, line, line_start, line_caret = \
-                self._get_text_after_prompt(accessible, offset)
+            self._get_text_after_prompt(accessible, offset)
         return bool(prompt_length)
 
     def can_suggest_before_typing(self):
@@ -597,17 +650,20 @@ class DomainTerminal(TextDomain):
 
         return True, None  # entering_text, end_of_editing
 
+    def can_auto_punctuate(self, has_begin_of_text):
+        """
+        Only auto-punctuate in Terminal when no prompt was detected.
+        Intention is to allow punctuation assistance in editors, but disable
+        it when entering commands at the prompt, e.g. for "cd ..".
+        """
+        return not has_begin_of_text
+
 
 class DomainURL(DomainGenericText):
     """ (Firefox) address bar """
 
-    def matches(self,  **kwargs):
-        attributes = kwargs.get("attributes")
-        if attributes:
-            # firefox url bar?
-            if "urlbar" in attributes.get("class", ""):
-                return True
-        return False
+    def matches(self, **kwargs):
+        return kwargs.get("is_urlbar", False)
 
     def get_auto_separator(self, context):
         """
@@ -618,7 +674,7 @@ class DomainURL(DomainGenericText):
     def get_text_begin_marker(self):
         return "<bot:url>"
 
-    def can_spell_check(self):
+    def can_spell_check(self, section_span):
         return False
 
 
@@ -680,7 +736,7 @@ class PartialURLParser:
     _protocols = ["mailto", "apt"]
     _all_schemes = _schemes + _protocols
 
-    _url_pattern = re.compile("([\w-]+)|(\W+)", re.UNICODE)
+    _url_pattern = re.compile(r"([\w-]+)|(\W+)", re.UNICODE)
 
     def iter_url(self, url):
         return self._url_pattern.finditer(url)
@@ -773,26 +829,36 @@ class PartialURLParser:
         ':'
 
         # local files
+        >>> import tempfile
+        >>> from os.path import abspath, dirname, join
+        >>> def touch(fn):
+        ...     with open(fn, mode="w") as f: n = f.write("")
+        >>> td = tempfile.TemporaryDirectory(prefix="test onboard _")
+        >>> dir = td.name
+
+        >>> touch(join(dir, "onboard-defaults.conf.example"))
+        >>> touch(join(dir, "onboard-defaults.conf.example.another"))
+        >>>
+        >>> import glob
+        >>> #glob.glob(dir+"/**")
         >>> p.get_auto_separator("file")
         ':///'
         >>> p.get_auto_separator("file:///home")
         '/'
-        >>> from os.path import abspath, dirname
-        >>> url = "file://" + dirname(abspath("./onboard-defaults.conf.example"))
-        >>> p.get_auto_separator(url + "/onboard-defaults")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults"))
         '.'
-        >>> p.get_auto_separator(url + "/onboard-defaults.conf")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults.conf"))
         '.'
-        >>> p.get_auto_separator(url + "/onboard-defaults.conf.example")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults.conf.example"))
         ''
-        >>> p.get_auto_separator(url + "/onboard-defaults.conf.example.nexus7")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults.conf.example.another"))
         ' '
 
         # Non-existing filename: we don't know, don't guess a separator
         >>> p.get_auto_separator("file:///tmp/onboard1234")
         ''
 
-        # Non-existing filename: if basename has an extension assumes we're done
+        # Non-existing filename: if basename has an extension assume we're done
         >>> p.get_auto_separator("file:///tmp/onboard1234.txt")
         ' '
 
@@ -874,7 +940,7 @@ class PartialURLParser:
 
         if os.path.isabs(filename):
             files  = glob.glob(filename + "*")
-            files += glob.glob(filename + "/*") # look inside directories too
+            files += glob.glob(filename + "/*")  # look inside directories too
             separator = self._get_separator_from_file_list(filename, files)
 
         if separator is None:
@@ -927,6 +993,14 @@ class PartialURLParser:
         '.'
         >>> p._get_separator_from_file_list("/dir/dir/file.ext1.ext2", files)
         ' '
+
+        # partial path match
+        >>> files = ["/dir/dir/file", "/dir/dir/file.ext1",
+        ...          "/dir/dir/file.ext2"]
+        >>> p._get_separator_from_file_list("/dir/dir/file", files)
+        ''
+        >>> p._get_separator_from_file_list("/dir/dir/file.ext1", files)
+        ' '
         """
         separator = None
 
